"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from libs import base
from copy import deepcopy
from controllers import cluster_controller
from constants import MULTICLUSTER_UPDATE_TIMEOUT
from datetime import datetime
from itertools import chain


class MultiCluster(object):
    def __init__(
            self, cluster_id, cluster_obj, cluster_name, xdr_port):
        """Creating multicluster object"""
        self._multicluster = {}
        self._discovered_cluster_objects = {}
        self._last_accessed = datetime.now()
        self._xdr_port = xdr_port
        self.add_cluster_object(
            cluster_id, cluster_obj, cluster_name, self._xdr_port)

    def __str__(self):
        return str(len(self._multicluster))

    def get_multicluster_data(
            self, xdr_port, session, session_manager, cluster_manager):
        """ Return multicluster info """
        self._last_accessed = datetime.now()

        """
        # Remove secured cluster if deleted from session (Logout)
        for cluster_id, cluster_obj in self._discovered_cluster_objects.items():
            response = session_manager.is_cluster_id_present(
                session['session'], cluster_id)
            if not response:
                # Remove only secured cluster so that user can login
                # with different credentials
                if cluster_obj.is_secured():
                    # Delete cluster from multicluster
                    del self._discovered_cluster_objects[cluster_id]
                    del self._multicluster[cluster_id]
                    # Refresh clusters containing the deleted cluster
                    for key, cluster_data in self._multicluster.items():
                        if cluster_id in cluster_data['xdr_info'].keys():
                            self.add_cluster_object(
                                key, self._discovered_cluster_objects[key],
                                cluster_data['cluster_name'], self._xdr_port
                            )
        """

        # If XDR port changes: update mc
        if xdr_port != self._xdr_port:
            self._xdr_port = xdr_port
            self.update_multicluster()

        # Create cluster objects of nodes that are available
        # Cannot add this step in background call as background call runs on
        # another thread and does not share global and local variables
        # giving error in creation of object
        temp_dict = deepcopy(self._multicluster)
        for cluster_id, cluster_info in temp_dict.items():
            if cluster_info['discovery'] is 'in_progress':
                ip, port = cluster_id.split(":")
                # Recheck availability
                node_available = base.get_node_availabilty(ip, port)
                if node_available == "available":
                    cluster_details = dict(seed_node=cluster_id)
                    # Create Cluster Object
                    response = cluster_controller.get_cluster_id(
                        cluster_details, session, session_manager,
                        cluster_manager
                    )
                    cluster_id = response.get('cluster_id', None)
                    if cluster_id:
                        cluster_obj = cluster_manager.get_cluster_by_id(
                            cluster_id)
                        self.add_cluster_object(
                            cluster_id, cluster_obj, None, self._xdr_port)
                else:
                    cluster_info['discovery'] = node_available
        del temp_dict

        response = {}
        for cluster_id, cluster_data in self._multicluster.items():
            response[cluster_id] = dict()
            for key in cluster_data.keys():
                if key in ['cluster_name', 'dc_name', 'nodes', 'xdr_info',
                           'discovery', 'read_tps', 'write_tps', 'seednode',
                           'namespaces']:
                    response[cluster_id][key] = cluster_data[key]
        return response

    def get_discovered_cluster_objects(self):
        """Return list of clusters who have cluster object"""
        return self._discovered_cluster_objects

    def add_cluster_object(self, cluster_id, cluster_obj,
                           cluster_name, xdr_port):
        """ Add cluster object to multicluster """
        # Get user credentials
        user_credentials = cluster_obj.get_user()
        if user_credentials:
            user = user_credentials.values()[0]['username']
            password = user_credentials.values()[0]['password']
        else:
            user, password = None, None

        # Prepare dict
        cluster_data = {
            cluster_id: dict(
                cluster_name=cluster_name,
                dc_name=[],
                xdr_port=self._xdr_port,
                cluster_key=None,
                cluster_size=None,
                read_tps=dict(total=0, success=0),
                write_tps=dict(total=0, success=0),
                namespaces=[],
                seednode=cluster_obj.get_seed(),
                nodes={},
                xdr_info={}
            )
        }
        node_obj_list = cluster_obj.get_node_id_node_obj()
        for node_id, node_obj in node_obj_list.items():
            ip = node_obj.get_ip()
            port = node_obj.get_port()
            status = node_obj.get_status()
            cluster_data[cluster_id]['nodes'][node_id] = dict(
                ip=ip, port=port, status=status, lag=None, cur_throughput=None,
                xdr_uptime=None)

            # Obtain namespaces once
            if not cluster_data[cluster_id]['namespaces']:
                cluster_data[cluster_id]['namespaces'] = list(
                    node_obj.get_namespace_list(user, password))

            if status is "on":
                # Obtain cluster size once
                if cluster_data[cluster_id]['cluster_size'] in [None, 'n/s']:
                    cluster_data[cluster_id][
                        'cluster_size'] = node_obj.get_statistic('cluster_size')
                # Obtain cluster key once
                if cluster_data[cluster_id]['cluster_key'] in [None, 'n/s']:
                    cluster_data[cluster_id][
                        'cluster_key'] = node_obj.get_statistic('cluster_key')
                # Get read/write tps cluster wide
                read_tps = node_obj.get_read_tps()
                if read_tps['secondary']:
                    cluster_data[
                        cluster_id][
                        'read_tps']['total'] += read_tps['secondary']
                    cluster_data[
                        cluster_id]['read_tps']['success'] += read_tps['y']

                write_tps = node_obj.get_write_tps()
                if write_tps['secondary']:
                    cluster_data[
                        cluster_id][
                            'write_tps']['total'] += write_tps['secondary']
                    cluster_data[
                        cluster_id]['write_tps']['success'] += write_tps['y']

            # Get XDR data of every node
            xdr_obj = node_obj.get_xdr(xdr_port, user=user, password=password)
            xdr_data = None
            if xdr_obj:
                xdr_stat = xdr_obj.get_subset_of_statistics(
                        ['timediff_lastship_cur_secs', 'cur_throughput',
                            'xdr-uptime'], user=user, password=password)[0]
                cluster_data[
                    cluster_id]['nodes'][node_id]['lag'] = xdr_stat[
                        'timediff_lastship_cur_secs']
                cluster_data[
                    cluster_id]['nodes'][node_id]['cur_throughput'] = xdr_stat[
                        'cur_throughput']
                # Get xdr uptime
                cluster_data[cluster_id]['nodes'][node_id][
                    'xdr_uptime'] = xdr_stat['xdr-uptime']

                xdr_data = xdr_obj.get_dc_config(user=user, password=password)

            if not xdr_data:
                continue
            # Add entry to xdr_info
            for ip, info in xdr_data.items():
                cluster_data[cluster_id]['xdr_info'][ip] = dict(
                    shipping_namespaces=info['namespaces'])
            # Create entry for that cluster
            self.add_xdr_cluster(xdr_data)

        cluster_data[cluster_id]['discovery'] = "complete"

        self.add_to_multicluster(cluster_data)
        self._discovered_cluster_objects[cluster_id] = cluster_obj

    def add_to_multicluster(self, cluster_data):
        """ Processing before inserting into multicluster
        {
            '<cluster_key/ip-address>': {
                'cluster_name': <cluster-name>,
                'dc_name': [<dc-name>, ],
                'namespaces': [<namespace>, ],
                'nodes': {
                    <node-id/node-ip>: {
                        'ip': <ip>,
                        'port': <port>,
                        'status': <on/off>,
                        'lag': <lag>,
                        'cur_throughput':<cur_throughput>
                    }
                },
                'xdr_info': {
                    <cluster_key/ip-address>:{
                        'shipping_namespaces': [<namespace>]
                    }
                }
                'discovery': <in_progress/complete/secured/not_available>
            }
        } """
        cluster_id = cluster_data.keys()[0]
        cluster_info = cluster_data.values()[0]

        for node_info in cluster_info['nodes'].values():
            node_address = ''.join(
                [node_info['ip'], ':', node_info['port']])
            # Check if this node_address is present in other cluster's
            # xdr_info, If yes then replace node_address with cluster_key
            for mc_id, mc_info in self._multicluster.items():
                for shipping_ip in mc_info['xdr_info'].keys():
                    if node_address == shipping_ip:
                        if cluster_id != shipping_ip:
                            # Get shipping namespaces
                            if cluster_id in mc_info['xdr_info'].keys():
                                mc_info['xdr_info'][cluster_id][
                                    'shipping_namespaces'] = list(
                                    set(
                                        mc_info['xdr_info'][
                                            cluster_id]['shipping_namespaces']
                                    ).union(
                                        mc_info['xdr_info'][
                                            shipping_ip]['shipping_namespaces']
                                    )
                                )
                                mc_info['xdr_info'].pop(shipping_ip)
                            else:
                                mc_info['xdr_info'][cluster_id] = mc_info[
                                    'xdr_info'].pop(shipping_ip)

                # Check if this node_address is present in multicluster
                # set, absorb dc names and then delete that entry
                if node_address == mc_id:
                    old_list = set(cluster_info['dc_name'])
                    new_list = mc_info['dc_name']
                    cluster_info['dc_name'] = list(old_list.union(new_list))
                    del self._multicluster[mc_id]

        if cluster_info['xdr_info']:
            for address, info in cluster_info['xdr_info'].items():
                # Check if this address is present in discovered
                # cluster's node list, If yes then replace
                # this entry with the cluster key of that node address
                if address in self._multicluster.keys():
                    continue
                changed = False
                for mc_id, mc_info in self._multicluster.items():
                    for node_info in mc_info['nodes'].values():
                        node = ''.join(
                            [node_info['ip'], ":", node_info['port']])
                        if address == node:
                            if mc_id != address:
                                cluster_info['xdr_info'][mc_id] = cluster_info[
                                    'xdr_info'].pop(address)
                                changed = True

                if not changed:
                    # Check if the cluster is shipping to itself
                    # This is a possible scenario
                    for nodes in cluster_info['nodes'].values():
                        node_address = ''.join(
                            [nodes['ip'], ':', nodes['port']])
                        if address == node_address:
                            if cluster_id != address:
                                cluster_info[
                                    'xdr_info'][cluster_id] = cluster_info[
                                        'xdr_info'].pop(address)

        self._multicluster.update(cluster_data)

    def add_xdr_cluster(self, xdr_data):
        """Add xdr entry to multicluster"""
        for node_address, cluster_info in xdr_data.items():
            already_present = False
            # If same entry already exists then just absorb dc_name
            if node_address in self._multicluster.keys():
                temp = set(self._multicluster[node_address]['dc_name'])
                temp.add(cluster_info['DC_Name'])
                self._multicluster[node_address]['dc_name'] = list(temp)
                already_present = True

            # Check if a cluster object representing same node exists
            for mc_info in self._multicluster.values():
                for address in mc_info['nodes'].values():
                    node = ''.join(
                        [address['ip'], ':', address['port']])
                    # Absorb dc_name
                    if node_address == node:
                        temp = set(mc_info['dc_name'])
                        temp.add(cluster_info['DC_Name'])
                        mc_info['dc_name'] = list(temp)
                        already_present = True

            if already_present:
                continue

            # Create fresh mc entry
            ip, port = node_address.split(":")
            node_available = base.get_node_availabilty(ip, port)

            if node_available in ['available', 'secured']:
                status = "on"
            else:
                status = "off"

            # Setting discovery of node which is available to in_progress
            # so that a cluster object can be created on next fetch for mc
            # if node_available is 'available':
            #     node_available = "in_progress"

            cluster_data = {
                node_address: dict(
                    cluster_name=None,
                    dc_name=[cluster_info['DC_Name'], ],
                    xdr_port=self._xdr_port,
                    cluster_key=None,
                    cluster_size=None,
                    read_tps=dict(total=0, success=0),
                    write_tps=dict(total=0, success=0),
                    namespaces=[],
                    seednode=node_address,
                    nodes={
                        node_address: dict(
                            ip=ip, port=port, status=status,
                            lag=None, cur_throughput=None, xdr_uptime=None
                        )
                    },
                    xdr_info={},
                    discovery=node_available
                )
            }
            self.add_to_multicluster(cluster_data)

    def update_multicluster(self):
        """ To update cluster info in multicluster """
        time_diff = datetime.now() - self._last_accessed
        # Stop updating after a timeout period
        if time_diff.seconds > MULTICLUSTER_UPDATE_TIMEOUT*60:
            return

        # Using deepcopy to avoid RuntimeError while changing size of dict
        # while iterating over it

        # Running update for multiclusters in sequence of independent
        # clusters, dependent clusters and clusters with no objects

        independent_clusters = {}
        dependent_clusters = {}
        non_object_clusters = {}

        for key in self._multicluster.keys():
            if key in self._discovered_cluster_objects.keys(
                ) and not self._discovered_cluster_objects[
                    key].get_mc_dependent():
                independent_clusters[key] = deepcopy(self._multicluster[key])

            elif key in self._discovered_cluster_objects.keys(
                ) and self._discovered_cluster_objects[
                    key].get_mc_dependent():
                dependent_clusters[key] = deepcopy(self._multicluster[key])
            elif key not in self._discovered_cluster_objects.keys():
                non_object_clusters[key] = deepcopy(self._multicluster[key])

        for cluster_id in chain(
                independent_clusters.keys(), dependent_clusters.keys(),
                non_object_clusters.keys()):
            cluster_obj = self._discovered_cluster_objects.get(
                cluster_id, None)
            try:
                cluster_info = self._multicluster[cluster_id]
            except KeyError:
                continue

            if cluster_obj:
                self.checks_for_cluster_with_objects(
                    cluster_id, cluster_obj, cluster_info)
            else:
                self.checks_for_cluster_without_object(cluster_id, cluster_info)
        del independent_clusters, dependent_clusters, non_object_clusters

    def checks_for_cluster_with_objects(
            self, cluster_id, cluster_obj, cluster_info):
        """Checks for cluster who have objects"""

        # If this cluster object is dependent and not being shipped to from
        # other cluster then do not show this entry
        deleted = False
        shipping = False
        if cluster_obj.get_mc_dependent():
            for info in self._multicluster.values():
                if cluster_id in info['xdr_info'].keys():
                    shipping = True
                    break
            if not shipping:
                del self._multicluster[cluster_id]
                del self._discovered_cluster_objects[cluster_id]
                deleted = True

        if deleted:
            return

        # Get user credentials
        user_credentials = cluster_obj.get_user()
        if user_credentials:
            user = user_credentials.values()[0]['username']
            password = user_credentials.values()[0]['password']
        else:
            user, password = None, None

        state_changed = False
        new_cluster_key = None
        new_cluster_size = None
        cluster_obj.update_nodes()
        node_obj_list = cluster_obj.get_node_id_node_obj()
        # msg = None

        if self._xdr_port != cluster_info['xdr_port']:
            # msg = "xdr port change"
            state_changed = True

        # Calculate read/write tps
        cluster_info['read_tps']['total'] = 0
        cluster_info['read_tps']['success'] = 0
        cluster_info['write_tps']['total'] = 0
        cluster_info['write_tps']['success'] = 0

        for node_id, node_obj in node_obj_list.items():
            # For xdr port change
            if state_changed:
                break
            if node_id not in cluster_info['nodes'].keys():
                state_changed = True
                # msg = "New Node added"
                break

            status = node_obj.get_status()
            if cluster_info[
                    'nodes'][node_id]['status'] != status:
                state_changed = True
                # msg = "Node status changed"
                break

            if status is "on":
                read_tps = node_obj.get_read_tps()
                if read_tps['secondary']:
                    cluster_info['read_tps']['total'] += read_tps['secondary']
                    cluster_info['read_tps']['success'] += read_tps['y']

                write_tps = node_obj.get_write_tps()
                if write_tps['secondary']:
                    cluster_info['write_tps']['total'] += write_tps['secondary']
                    cluster_info['write_tps']['success'] += write_tps['y']

                if new_cluster_size in [None, 'n/s']:
                    # Check once
                    new_cluster_size = node_obj.get_statistic('cluster_size')
                    if cluster_info['cluster_size'] in [None, 'n/s']:
                        cluster_info['cluster_size'] = new_cluster_size
                    elif cluster_info['cluster_size'] != new_cluster_size:
                        # msg = "Node cluster size changed"
                        state_changed = True
                        break

                if new_cluster_key in [None, 'n/s']:
                    new_cluster_key = node_obj.get_statistic('cluster_key')
                    if cluster_info['cluster_key'] in [None, 'n/s']:
                        cluster_info['cluster_key'] = new_cluster_key
                    elif cluster_info['cluster_key'] != new_cluster_key:
                        # msg = "Cluster key changed"
                        state_changed = True
                        break

                # Get namespaces if null
                if not cluster_info['namespaces']:
                    cluster_info['namespaces'] = list(
                        node_obj.get_namespace_list(user, password))

            xdr_obj = node_obj.get_xdr(cluster_info['xdr_port'])
            if xdr_obj:
                xdr_stat = xdr_obj.get_subset_of_statistics(
                        ['timediff_lastship_cur_secs', 'cur_throughput',
                         'xdr-uptime'], user=user, password=password)[0]

                if xdr_stat['xdr-uptime'] in [None, 'n/s']:
                    xdr_availability = False
                else:
                    xdr_availability = True

                old_xdr_uptime = cluster_info['nodes'][node_id]['xdr_uptime']
                if old_xdr_uptime not in [None, 'n/s'] and \
                        not xdr_availability:
                    state_changed = True
                    # msg = "XDR Inactive"
                    break
                elif old_xdr_uptime in [None, 'n/s'] and xdr_availability:
                    state_changed = True
                    # msg = "XDR Active"
                    break

                # Check if XDR was restarted
                if xdr_stat['xdr-uptime'] not in [None, 'n/s']:
                    try:
                        if int(xdr_stat['xdr-uptime']) < 5:
                            state_changed = True
                            # msg = "XDR restarted"
                            break
                    except ValueError:
                        pass
                cluster_info['nodes'][node_id][
                    'xdr_uptime'] = xdr_stat['xdr-uptime']

                # If previous values of lag and cur_throughput are not set then
                # set them
                if cluster_info['nodes'][node_id]['lag'] in [
                        None, 'n/s'] or cluster_info['nodes'][node_id][
                            'cur_throughput'] in [None, 'n/s']:
                        cluster_info['nodes'][node_id][
                            'lag'] = xdr_stat[
                                'timediff_lastship_cur_secs']
                        cluster_info['nodes'][node_id][
                            'cur_throughput'] = xdr_stat['cur_throughput']
                # If previous values are set and do not match with the current
                # values then state change
                elif cluster_info['nodes'][node_id]['lag'] != xdr_stat[
                        'timediff_lastship_cur_secs'] or cluster_info[
                            'nodes'][node_id]['cur_throughput'] != xdr_stat[
                                'cur_throughput']:
                        # msg = "Node XDR stat change"
                        state_changed = True
                        break

        if state_changed:
            # sys.stderr.write("\n[INFO] State changed of %s Reason: %s" % (
            #     cluster_obj.get_seed(), msg))

            self.add_cluster_object(
                cluster_id, cluster_obj, cluster_info['cluster_name'],
                self._xdr_port
            )

    def checks_for_cluster_without_object(
            self, cluster_id, cluster_info):
        """Checks for cluster without object"""
        # Check if this cluster is being referenced by any cluster having
        # object , If not then delete this entry
        present = False
        for mc_id, mc_info in self._multicluster.items():
            if mc_id == cluster_id:
                continue

            if mc_id not in self._discovered_cluster_objects.keys():
                continue

            if cluster_id in mc_info['xdr_info'].keys():
                present = True
                break

        if not present:
            del self._multicluster[cluster_id]
            return

        # Mark available node for creating object
        if cluster_info['discovery'] == 'available':
            cluster_info['discovery'] = 'in_progress'
        else:
            ip, port = cluster_id.split(":")
            node_available = base.get_node_availabilty(ip, port)
            if node_available is 'available':
                cluster_info['discovery'] = "in_progress"
